package info.textgrid.middleware;

import static org.junit.Assert.assertTrue;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.ws.rs.core.Response;
import org.apache.cxf.helpers.IOUtils;
import org.apache.cxf.jaxrs.client.Client;
import org.apache.cxf.jaxrs.client.JAXRSClientFactory;
import org.apache.cxf.jaxrs.client.WebClient;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;
import org.apache.http.HttpStatus;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

/**
 * <p>
 * Some online tests for the TextGrid OAIMPH service.
 * </p>
 * 
 * @author Stefan E. Funk, SUB Göttingen
 */
//@Ignore
public class OaiPmhTextgridOnlineTests {

  // **
  // FINALS
  // **

  private static final String PROPERTIES_FILE = "oaipmh.test.textgridlab-org.properties";
  // NOTE/TODO There is NO IDIOM test data for testing on dev.textgridlab.org! So eventually all
  // IDIOM tests will fail!
  // private static final String PROPERTIES_FILE = "oaipmh.test.dev-textgridlab-org.properties";

  // **
  // STATICS
  // **

  // Some JAXRS things.
  private static String oaipmhEndpoint;
  private static OAIPMHProducer JAXRSClient;
  static Client oaipmhWebClient;

  // Properties
  private static String expectedListSets;
  private static String checkGetRecordDC;
  private static String expectedGetRecordDC;
  private static String checkGetRecordIDIOM;
  private static String checkGetRecordIDIOM_Images;
  private static String expectedGetRecordIDIOM_Images;
  private static String expectedGetRecordIDIOM;
  private static String checkGetRecordDATACITE;
  private static String expectedGetRecordDATACITE;
  private static String checkListRecordsDC;
  private static String checkListRecordsDCFrom;
  private static String checkListRecordsDCUntil;
  private static String checkListRecordsIDIOMFrom;
  private static String checkListRecordsIDIOMUntil;
  private static String checkListRecordsDATACITEFrom;
  private static String checkListRecordsDATACITEUntil;
  private static String checkListIdentifiersSet;
  private static int checkListIdentifiersPagesToTestIDIOM;
  private static int checkListIdentifiersRecordsPerPageIDIOM;
  private static List<String> checkGetRecordIDList = new ArrayList<String>();

  // **
  // PREPARATIONS
  // **

  /**
   * @throws Exception
   */
  @BeforeClass
  public static void setUpBeforeClass() throws Exception {

    // Load properties file.
    Properties p = new Properties();
    p.load(new FileInputStream(OaiPmhTestUtilities.getResource(PROPERTIES_FILE)));

    System.out.println("Properties file: " + PROPERTIES_FILE);
    p.list(System.out);

    // Get other needed properties.
    oaipmhEndpoint = p.getProperty("oaipmhEndpoint");
    expectedListSets = p.getProperty("expectedListSets");
    checkGetRecordDC = p.getProperty("checkGetRecordDC");
    expectedGetRecordDC = p.getProperty("expectedGetRecordDC");
    checkGetRecordIDIOM = p.getProperty("checkGetRecordIDIOM");
    checkGetRecordIDIOM_Images = p.getProperty("checkGetRecordIDIOM_Images");
    expectedGetRecordIDIOM_Images = p.getProperty("expectedGetRecordIDIOM_Images");
    expectedGetRecordIDIOM = p.getProperty("expectedGetRecordIDIOM");
    checkGetRecordDATACITE = p.getProperty("checkGetRecordDATACITE");
    expectedGetRecordDATACITE = p.getProperty("expectedGetRecordDATACITE");
    checkListRecordsDC = p.getProperty("checkListRecordsDC");
    checkListRecordsDCFrom = p.getProperty("checkListRecordsDCFrom");
    checkListRecordsDCUntil = p.getProperty("checkListRecordsDCUntil");
    checkListRecordsIDIOMFrom = p.getProperty("checkListRecordsIDIOMFrom");
    checkListRecordsIDIOMUntil = p.getProperty("checkListRecordsIDIOMUntil");
    checkListRecordsDATACITEFrom = p.getProperty("checkListRecordsDATACITEFrom");
    checkListRecordsDATACITEUntil = p.getProperty("checkListRecordsDATACITEUntil");
    checkListIdentifiersSet = p.getProperty("checkListIdentifiersSet");
    checkListIdentifiersPagesToTestIDIOM =
        Integer.parseInt(p.getProperty("checkListIdentifiersPagesToTestIDIOM"));
    checkListIdentifiersRecordsPerPageIDIOM =
        Integer.parseInt(p.getProperty("checkListIdentifiersRecordsPerPageIDIOM"));
    checkGetRecordIDList =
        OaiPmhTestUtilities.getListFromProperties((String) p.get("checkGetRecordIDList"));

    // Get OAI-PMH REST endpoint and HTTP client.
    System.out.println("Getting OAI-PMH HTTP client --> " + oaipmhEndpoint + " <--");

    // Get proxy first, set policy.
    JAXRSClient = JAXRSClientFactory.create(oaipmhEndpoint, OAIPMHProducer.class);
    HTTPConduit conduit = WebClient.getConfig(JAXRSClient).getHttpConduit();
    HTTPClientPolicy policy = new HTTPClientPolicy();
    policy.setReceiveTimeout(OaiPmhTestUtilities.OAIPMH_CLIENT_TIMEOUT);
    conduit.setClient(policy);

    // Create Web Client from Web Proxy.
    oaipmhWebClient = WebClient.client(JAXRSClient);
  }

  /**
   * @throws Exception
   */
  @AfterClass
  public static void tearDownAfterClass() throws Exception {
    //
  }

  /**
   * @throws Exception
   */
  @Before
  public void setUp() throws Exception {
    //
  }

  /**
   * @throws Exception
   */
  @After
  public void tearDown() throws Exception {
    //
  }

  // **
  // TESTS
  // **

  /**
   * @throws IOException
   */
  @Test
  public void testGetVersion() throws IOException {

    String shouldStartWith = "oaipmh-core";

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetVersion()");

    String versionString = JAXRSClient.getVersion();

    if (!versionString.startsWith(shouldStartWith)) {
      System.err.println(
          OaiPmhTestUtilities.ERROR + ": response should start with '" + shouldStartWith + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + versionString);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testRootUrl() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRootUrl()");

    // Still hard coded in OAIPMH service!
    String shouldContain = "textgridrep.org";

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, "oai",
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testIdentify() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testIdentify()");

    // Still hard coded in OAIPMH service!
    String shouldContain = "textgridrep.org";

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient,
        "verb=" + OaiPmhTestUtilities.VERB_IDENTIFY, OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListMetadataFormats() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListMetadataFormats()");

    String shouldContainDC = OaiPmhTestUtilities.OAI_DC_PREFIX;
    String shouldContainIDIOM = OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX;
    String shouldContainDATACITE = OaiPmhTestUtilities.OAI_DATACITE_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient,
        "verb=" + OaiPmhTestUtilities.VERB_LIST_METADATA_FORMATS,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContainDC)
        || !response.contains(shouldContainIDIOM) || !response.contains(shouldContainDATACITE)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContainDC
          + "' and '" + shouldContainIDIOM + "' and '" + shouldContainDATACITE + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListSets() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListSets()");

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient,
        "verb=" + OaiPmhTestUtilities.VERB_LIST_SETS, OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(expectedListSets)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '"
          + expectedListSets + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordOAIDC() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordOAIDC()");

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + checkGetRecordDC + "&metadataPrefix="
        + OaiPmhTestUtilities.OAI_DC_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDC)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '"
          + expectedGetRecordDC + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordOAIDCNoPrefix() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordOAIDCNoPrefix()");

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + checkGetRecordDC + "&metadataPrefix="
        + OaiPmhTestUtilities.OAI_DC_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDC)) {
      System.err.println("\tstatus: " + status);
      System.err.println(
          OaiPmhTestUtilities.ERROR + ": response should contain '"
              + checkGetRecordDC + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETS() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordIDIOMMETS()");

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + checkGetRecordIDIOM + "&metadataPrefix="
        + OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordIDIOM)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '"
          + expectedGetRecordIDIOM + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETS_Image() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordIDIOMMETS_Image()");

    String path =
        "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier=" + checkGetRecordIDIOM_Images
            + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordIDIOM_Images)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '"
          + expectedGetRecordIDIOM_Images + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETSNoPrefix() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordIDIOMMETSNoPrefix()");

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + checkGetRecordIDIOM + "&metadataPrefix="
        + OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordIDIOM)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '"
          + expectedGetRecordIDIOM + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordDATACITE() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordDATACITE()");

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + checkGetRecordDATACITE + "&metadataPrefix="
        + OaiPmhTestUtilities.OAI_DATACITE_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDATACITE)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '"
          + expectedGetRecordDATACITE + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordDATACITENoPrefix() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordDATACITENoPrefix()");

    String path =
        "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
            + checkGetRecordDATACITE
            + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_DATACITE_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();
    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDATACITE)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain '"
          + expectedGetRecordDATACITE + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsOAIDCMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsOAIDCMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, checkListRecordsDC,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 20, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsOAIDCMorePagesFromUntil() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsOAIDCMorePagesFromUntil()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, checkListRecordsDC,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 0, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        checkListRecordsDCFrom, checkListRecordsDCUntil,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test(expected = IOException.class)
  public void testListRecordsIDIOMMETSSomePagesAndMetadataFormatWithRestok() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING
        + "testListRecordsIDIOMMETSSomePagesAndMetadataFormatWithRestok()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  @Ignore
  public void testListRecordsIDIOMMETSAllPages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsIDIOMMETSAllPages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 0, 30, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * NOTE The amount of records will change due to changes on the data! Please check from and until
   * values before!
   *
   * @throws IOException
   */
  @Test
  @Ignore
  public void testListRecordsIDIOMMETSMorePagesFromUntil() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsIDIOMMETSMorePagesFromUntil()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 6, 30, OaiPmhTestUtilities.NO_THREAD_NAME,
        checkListRecordsIDIOMFrom,
        checkListRecordsIDIOMUntil,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsIDIOMMETSMorePages() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsIDIOMMETSMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 6, 30, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsDATACITEMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsDATACITEMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 15, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  @Ignore
  public void testListRecordsDATACITEAllPages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsDATACITEAllPages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 0, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsDATACITEMorePagesFromUntil() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsDATACITEMorePagesFromUntil()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 15, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        checkListRecordsDATACITEFrom,
        checkListRecordsDATACITEUntil,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersOAIDCMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersOAIDCMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        checkListIdentifiersSet, OaiPmhTestUtilities.OAI_DC_PREFIX, 30,
        100,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test(expected = IOException.class)
  public void testListIdentifiersOAIDCMorePagesAndMetadataFormatWithRestok() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING
        + "testListIdentifiersOAIDCMorePagesAndMetadataFormatWithRestok()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        checkListIdentifiersSet, OaiPmhTestUtilities.OAI_DC_PREFIX, 30,
        100,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersIDIOMMETSAllPages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersIDIOMMETSAllPages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 0, 30,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersIDIOMMETSMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersIDIOMMETSMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 6, 30,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test(expected = IOException.class)
  public void testListIdentifiersIDIOMMETSSomePagesAndMetadataFormatWithRestok()
      throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING
        + "testListIdentifiersIDIOMMETSSomePagesAndMetadataFormatWithRestok()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersDATACITEMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersDATACITEMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 22, 100,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   * @throws ExecutionException
   * @throws InterruptedException
   */
  @Test
  public void testListIdentifiersConcurrentlyIDIOMMETSMorePages()
      throws IOException, InterruptedException, ExecutionException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "testListIdentifiersConcurrentlyIDIOMMETSMorePages()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30, "C1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 5, 30, "C2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 8, 30, "C2"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [C1]=" + f1.get() + ", [C2]=" + f2.get()
        + ", [C3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws ExecutionException
   * @throws InterruptedException
   */
  @Test
  public void testListRecordsConcurrentlyIDIOMMETSMorePages()
      throws IOException, InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsConcurrentlyIDIOMMETSMorePages()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 2, 30, "D1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30, "D2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 4, 30, "D3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [D1]=" + f1.get() + ", [D2]=" + f2.get()
        + ", [D3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersDC()
      throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersDC()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 36, 100, "A1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 27, 100, "A2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 19, 100, "A3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get()
        + ", [A3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersIDIOM()
      throws InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersIDIOM()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 13, 30, "A1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 35, 30, "A2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 11, 30, "A3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get()
        + ", [A3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersDATACITE()
      throws InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersDATACITE()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 13, 100, "A1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 20, 100, "A2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 8, 100, "A3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get()
        + ", [A3]=" + f3.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersDCAndIDIOMMets()
      throws InterruptedException, ExecutionException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersDCAndIDIOMMets()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 53, 100, "DC1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 28, 100, "DC2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX,
            checkListIdentifiersPagesToTestIDIOM,
            checkListIdentifiersRecordsPerPageIDIOM,
            "IDIOM"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [DC1]=" + f1.get() + ", [DC2]=" + f2.get()
        + ", [IDIOM]=" + f3.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListRecordsDC()
      throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListRecordsDC()");

    ExecutorService executor = Executors.newFixedThreadPool(4);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 44, 100, "DC1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 33, 100, "DC2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 66, 100, "DC3"));
    Future<Boolean> f4 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 28, 100, "DC4"));

    executor.shutdown();

    System.out.println(
        OaiPmhTestUtilities.OK + ": [DC1]=" + f1.get() + ", [DC2]=" + f2.get() + ", [DC3]="
            + f3.get() + ", [DC4]=" + f4.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListRecords() throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListRecords()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 6, 100, "B1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 7, 100, "B2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 8, 100, "B3"));
    Future<Boolean> f4 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 9, 100, "B4"));

    executor.shutdown();

    System.out.println(
        OaiPmhTestUtilities.OK + ": [B1]=" + f1.get() + ", [B2]=" + f2.get() + ", [B3]=" + f3.get()
            + ", [B4]=" + f4.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testListRecordsRestokConcurrentlyDCAndIDIOMMets()
      throws InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsRestokConcurrentlyDCAndIDIOMMets()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 34, 100, "DC1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 27, 100, "DC2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX,
            checkListIdentifiersPagesToTestIDIOM,
            checkListIdentifiersRecordsPerPageIDIOM,
            "IDIOM"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [DC1]=" + f1.get() + ", [DC2]=" + f2.get()
        + ", [IOIOM]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testConcurrentlyOAIDCGetRecord() throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testConcurrentlyOAIDCGetRecord()");

    ExecutorService executor = Executors.newFixedThreadPool(18);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(0),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(1),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(2),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f4 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(3),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f5 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(4),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f6 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(5),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f7 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(6),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f8 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(7),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f9 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(8),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f10 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(9),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f11 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(10),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f12 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(11),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f13 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(12),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f14 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(13),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f15 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(14),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f16 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(15),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f17 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(16),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f18 = executor
        .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD,
            checkGetRecordIDList.get(17),
            OaiPmhTestUtilities.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + "[GR1]=" + f1.get() + ", [GR2]=" + f2.get()
        + ", [GR3]=" + f3.get() + ", [GR4]=" + f4.get() + ", [GR5]=" + f5.get() + ", [GR6]="
        + f6.get() + ", [GR7]=" + f7.get() + ", [GR8]=" + f8.get() + ", [GR9]=" + f9.get()
        + ", [GR10]=" + f10.get() + ", [GR11]=" + f11.get() + ", [GR12]=" + f12.get()
        + ", [GR13]=" + f13.get() + ", [GR14]=" + f14.get() + ", [GR15]=" + f15.get()
        + ", [GR16]=" + f16.get() + ", [GR17]=" + f17.get() + ", [GR18]=" + f18.get());
  }

  /**
   * <p>
   * Starts a new OAI-PMH thread.
   * </p>
   */

  private class OaiPmhResumptionTokenTestThread implements Callable<Boolean> {

    protected String verb;
    protected String set;
    protected String metadataPrefix;
    protected int numberOfPagesToTest;
    protected int recordsExpectedPerRequest;
    protected String threadName;

    /**
     * @param theVerb
     * @param theSet
     * @param theMetadataPrefix
     * @param numberOfPagesToTest
     * @param recordsExpectedPerRequest
     */
    public OaiPmhResumptionTokenTestThread(String theVerb, String theSet, String theMetadataPrefix,
        int numberOfPagesToTest, int recordsExpectedPerRequest, String theThreadName) {
      this.verb = theVerb;
      this.set = theSet;
      this.metadataPrefix = theMetadataPrefix;
      this.numberOfPagesToTest = numberOfPagesToTest;
      this.recordsExpectedPerRequest = recordsExpectedPerRequest;
      this.threadName = theThreadName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public Boolean call() throws IOException {
      testList(oaipmhWebClient, this.verb, this.set, this.metadataPrefix,
          this.numberOfPagesToTest, this.recordsExpectedPerRequest, this.threadName,
          OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
          OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);
      return true;
    }
  }

  /**
   * <p>
   * Starts a new OAI-PMH thread.
   * </p>
   */

  private class OaiPmhGetRecordTestThread implements Callable<Boolean> {

    protected String verb;
    protected String identifier;
    protected String metadataPrefix;
    protected String threadName;

    /**
     * @param theVerb
     * @param theIdentifier
     * @param theMetadataPrefix
     * @param theThreadName
     */
    public OaiPmhGetRecordTestThread(String theVerb, String theIdentifier, String theMetadataPrefix,
        String theThreadName) {
      this.verb = theVerb;
      this.identifier = theIdentifier;
      this.metadataPrefix = theMetadataPrefix;
      this.threadName = theThreadName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.concurrent.Callable#call()
     */
    @Override
    public Boolean call() throws IOException {

      String url = "verb=" + this.verb + "&identifier=" + this.identifier + "&metadataPrefix="
          + this.metadataPrefix;
      OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, url, this.threadName);

      return true;
    }
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theClient
   * @param theVerb
   * @param theSet
   * @param theMetadataPrefix
   * @param maxNumberOfPagesToTest
   * @param recordsExpectedPerRequest
   * @param theThreadName
   * @param from
   * @param until
   * @param resumptionTokenANDMetadataPrefix
   * @throws IOException
   */
  protected static void testList(Client theClient, final String theVerb, final String theSet,
      final String theMetadataPrefix, final int maxNumberOfPagesToTest,
      final int recordsExpectedPerRequest, final String theThreadName, final String from,
      final String until, final boolean resumptionTokenANDMetadataPrefix) throws IOException {

    String threadName = tn(theThreadName);

    long startTime = System.currentTimeMillis();

    String testOccurance = "header";
    if (theVerb.equals(OaiPmhTestUtilities.VERB_LIST_RECORDS)) {
      testOccurance = "record";
    }

    String path = "verb=" + theVerb + "&metadataPrefix=" + theMetadataPrefix;

    if (theSet != null) {
      path += "&set=" + theSet;
    }

    if (from != null && until != null) {
      path += "&from=" + from + "&until=" + until;
    }

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(theClient, path, threadName);
    int status = httpResponse.getStatus();

    int loopCount = 1;
    long timeRunning = System.currentTimeMillis() - startTime;
    System.out.println("\t" + threadName + "time: "
        + OaiPmhTestUtilities.getDurationInSecs(timeRunning) + " (loop " + loopCount
        + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

    String responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    // Test resumption token tags.
    String restok = examineResumptionTokenTag(responseString, testOccurance,
        OaiPmhTestUtilities.NO_TOKEN, recordsExpectedPerRequest, loopCount, threadName);

    // Test general metadata content (if verb is listRecords!), must go conform with the metadata
    // prefix setting.
    if (theVerb.equals(OaiPmhTestUtilities.VERB_LIST_RECORDS)) {
      examineContent(responseString, theMetadataPrefix);
    }

    while (status == HttpStatus.SC_OK && !restok.equals(OaiPmhTestUtilities.NO_TOKEN)) {
      loopCount += 1;
      synchronized (threadName) {
        if (maxNumberOfPagesToTest > 0 && loopCount > maxNumberOfPagesToTest) {
          System.out
              .println("\t" + threadName + "TESTING ONLY " + maxNumberOfPagesToTest + " pages!");
          break;
        }
        path = "verb=" + theVerb + "&resumptionToken=" + restok;
        if (resumptionTokenANDMetadataPrefix) {
          path += "&metadataPrefix=" + theMetadataPrefix;
        }
        httpResponse = OaiPmhTestUtilities.getHttpResponse(theClient, path, threadName);
        timeRunning = System.currentTimeMillis() - startTime;
        System.out.println("\t" + threadName + "time: "
            + OaiPmhTestUtilities.getDurationInSecs(timeRunning) + " (loop " + loopCount
            + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

        // Test resumption token tags.
        responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());
        restok = examineResumptionTokenTag(responseString, testOccurance, restok,
            recordsExpectedPerRequest, loopCount, threadName);

        // Test general metadata content (if verb is listRecords!), must go conform with the
        // metadata // prefix setting.
        if (theVerb.equals(OaiPmhTestUtilities.VERB_LIST_RECORDS)) {
          examineContent(responseString, theMetadataPrefix);
        }
      }
    }

    // Only check for max loops, if we do have less loops and no resumption token, it will be fine,
    // too!
    if (!restok.equals(OaiPmhTestUtilities.NO_TOKEN) && loopCount < maxNumberOfPagesToTest) {
      System.out.println(threadName + OaiPmhTestUtilities.ERROR + ": Must have done "
          + maxNumberOfPagesToTest + " loops, but did only " + loopCount);
      assertTrue(false);
    }

    // Check for single loop, we need MORE!
    if (loopCount == 1) {
      System.out
          .println(threadName + OaiPmhTestUtilities.ERROR + ": Must have more than one loop!");
      assertTrue(false);
    }

    System.out.println("\t" + threadName + OaiPmhTestUtilities.OK);
  }

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theResponse
   * @param recordOrHeader
   * @param oldtok
   * @param recordsExpectedPerRequest
   * @param loopCount
   * @param theThreadName
   * @return Resumption token string, if existing, -1 if either no <resumptionToken> tag is
   *         existing, tag is existing and has no token value.
   * @throws IOException
   */
  public static String examineResumptionTokenTag(final String theResponseString,
      final String recordOrHeader, final String oldtok, final int recordsExpectedPerRequest,
      final int loopCount, final String theThreadName) throws IOException {

    // Test for OAIPMH errors.
    if (theResponseString.contains("<error code=\"badArgument\">")) {
      String message =
          theThreadName + OaiPmhTestUtilities.ERROR + " IN OAIPMH RESPONSE: " + theResponseString;
      System.out.println(message);
      throw new IOException(message);
    }

    // Count response objects at first.
    int recordCount = 0;
    int i = theResponseString.indexOf("<" + recordOrHeader + ">", 0);
    while (i != -1) {
      recordCount++;
      i++;
      i = theResponseString.indexOf("<" + recordOrHeader + ">", i);
    }

    System.out.println("\t" + theThreadName + recordOrHeader + "s: " + recordCount);

    // Check if token tag is existing.
    int tokStart = theResponseString.indexOf("<resumptionToken");
    int tokEnd = theResponseString.indexOf("</resumptionToken");

    if (tokStart == -1 && tokEnd == -1) {
      System.out.println("\t" + theThreadName + "token: no token");

      return OaiPmhTestUtilities.NO_TOKEN;
    }

    String restokTmp = theResponseString.substring(tokStart, tokEnd);
    // Get token tag.
    String toktag = restokTmp.substring(0, restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "tokentag: " + toktag);

    // Get token.
    String restok = restokTmp.substring(restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "token: " + restok);

    // Check if old and new token are equal or not.
    boolean tokchanged = !oldtok.equals(restok);
    System.out.println("\t" + theThreadName + "tokchngd: " + tokchanged);

    // Get completeListSize and cursor.
    String sizeStr = toktag.substring(toktag.indexOf("completeListSize=\"") + 18);
    int size = Integer.parseInt(sizeStr.substring(0, sizeStr.indexOf("\"")));
    String cursorStr = toktag.substring(toktag.indexOf("cursor=\"") + 8);
    int cursor = Integer.parseInt(cursorStr.substring(0, cursorStr.indexOf("\"")));
    System.out.println("\t" + theThreadName + "size: " + size + " / " + cursor);

    // If token is provided, test cursor and element count!
    if (!restok.isEmpty()) {
      synchronized (OaiPmhTextgridOnlineTests.class) {
        // Check <record> or <header> count, must be recordsExpectedPerRequest!
        if (recordCount != recordsExpectedPerRequest) {
          String message =
              OaiPmhTestUtilities.ERROR + ": " + recordOrHeader + " count mismatch, must be "
                  + recordsExpectedPerRequest + " if token is provided, but is " + recordCount;
          assertTrue(message, false);
        }
        if (size <= recordsExpectedPerRequest) {
          String message =
              OaiPmhTestUtilities.ERROR + ": completeListSize count mismatch, must be > "
                  + recordsExpectedPerRequest + " if token is provided, but is " + size;
          assertTrue(message, false);
        }
        if (cursor != recordsExpectedPerRequest * loopCount) {
          String message = OaiPmhTestUtilities.ERROR + ": cursor must be "
              + (loopCount * recordsExpectedPerRequest) + " in loop " + loopCount + ", but is "
              + cursor;
          assertTrue(message, false);
        }
      }
    }

    // If no token is provided, stop querying.
    else {
      // Check <record> count, must be completeListSize % recordsExpectedPerRequest.
      if (recordCount != size % recordsExpectedPerRequest) {
        System.err.println(theThreadName + OaiPmhTestUtilities.ERROR + ": " + recordOrHeader
            + " count mismatch, should be " + size % recordsExpectedPerRequest + ", but is "
            + recordCount);
        assertTrue(false);
      }

      // No resumption token available in response.
      return OaiPmhTestUtilities.NO_TOKEN;
    }

    return restok;
  }

  /**
   * @param theThreadName
   * @return
   */
  private static String tn(String theThreadName) {

    String result = theThreadName;

    if (theThreadName != null && !theThreadName.equals("") && !theThreadName.endsWith(" ")) {
      result = "[" + theThreadName + "] ";
    }

    return result;
  }

  /**
   * @param theResponseString
   * @param theMetadataFormat
   */
  private static void examineContent(String theResponseString, String theMetadataFormat) {

    // Check for correct metadata content according to metadata prefix.
    if (theMetadataFormat.equals(OaiPmhTestUtilities.OAI_DC_PREFIX)
        && !theResponseString
            .contains("metadataPrefix=\"" + OaiPmhTestUtilities.OAI_DC_PREFIX + "\"")
        && !theResponseString.contains(OaiPmhTestUtilities.EXPECTED_OAIDC_FORMAT_CONTENT)) {
      System.out
          .println(OaiPmhTestUtilities.OAI_DC_PREFIX + " needs to deliver content with schema: "
              + OaiPmhTestUtilities.EXPECTED_OAIDC_FORMAT_CONTENT + "!");
      assertTrue(false);
    } else if (theMetadataFormat.equals(OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX)
        && !theResponseString.contains(
            "metadataPrefix=\"" + OaiPmhTestUtilities.EXPECTED_IDIOMMETS_FORMAT_CONTENT + "\"")
        && !theResponseString.contains(OaiPmhTestUtilities.EXPECTED_IDIOMMETS_FORMAT_CONTENT)) {
      System.out.println(
          OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX + " needs to deliver content with schema: "
              + OaiPmhTestUtilities.EXPECTED_IDIOMMETS_FORMAT_CONTENT + "!");
      assertTrue(false);
    } else if (theMetadataFormat.equals(OaiPmhTestUtilities.OAI_DATACITE_PREFIX)
        && !theResponseString
            .contains("metadataPrefix=\"" + OaiPmhTestUtilities.OAI_DATACITE_PREFIX + "\"")
        && !theResponseString.contains(OaiPmhTestUtilities.EXPECTED_DATACITE_FORMAT_CONTENT)) {
      System.out.println(
          OaiPmhTestUtilities.OAI_DATACITE_PREFIX + " needs to deliver content with schema: "
              + OaiPmhTestUtilities.EXPECTED_DATACITE_FORMAT_CONTENT + "!");
      assertTrue(false);
    }
  }

}
